#include "G2D.h"

#ifdef __linux__ 
#include "GL/glut.h"
#endif

#ifdef  _WIN32
#include "glut.h" 
#endif


#ifdef __APPLE__
#include "GLUT/glut.h"
#endif


#include <string>
#include <map>
#include <vector>
#include <chrono>
 

using namespace std;


const Color Color::Black   = Color(0, 0, 0);
const Color Color::White   = Color(1, 1, 1);
const Color Color::Red     = Color(1, 0, 0);
const Color Color::Green   = Color(0, 1, 0);
const Color Color::Blue    = Color(0, 0, 1);
const Color Color::Magenta = Color(1, 0, 1);
const Color Color::Cyan    = Color(0, 1, 1);
const Color Color::Yellow  = Color(1, 1, 0);
const Color Color::Gray    = Color(0.5, 0.5, 0.5);


/////////////////////////////////////////////////////////////
//
//	    Create Texture From String
//
/////////////////////////////////////////////////////////////


typedef unsigned char u8;

u8 LetterToColor[256][4];

void AssignColor(int code,u8 R, u8 G, u8 B, u8 A=255)
{

	LetterToColor[code][0] = R;
	LetterToColor[code][1] = G;
	LetterToColor[code][2] = B;
	LetterToColor[code][3] = A;
}

void InitColors()
{
	// https://www.rapidtables.org/fr/web/color/RGB_Color.html

	AssignColor('R', 255, 0, 0);
	AssignColor('G', 0, 255, 0);
	AssignColor('B', 0, 0, 255);
	AssignColor('K', 0, 0, 0); // noir
	AssignColor('W', 255, 255, 255); // white
	AssignColor('M', 255, 0, 255); // magenta
	AssignColor('C', 0, 255, 255); // cyan
	AssignColor('Y', 255, 255, 0); // yellow
	AssignColor('S', 192, 192, 192); // silver
	AssignColor('G', 128, 128, 128); // gray
	AssignColor('O', 255, 215, 0); // or
	AssignColor(' ', 255, 255, 255, 0); // transparent
	AssignColor('[', 255, 255, 255, 0); // transparent
	AssignColor(']', 255, 255, 255, 0); // transparent
}

int CORECreateTextureFromRGBA(unsigned char * data, int width, int height)
{
	GLuint t = 0;
	glGenTextures(1, &t);
	glBindTexture(GL_TEXTURE_2D, t);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
	return t;
}

int IdStripeTexture;

 
void InitStripeTexture()
{
	unsigned char data[32 * 32 * 4];
	int nb = 32 * 32 * 4;
	for (int i = 0; i < nb; i += 8) { data[i] = data[i + 1] = data[i + 2] = 0;  data[i + 3] = 255; } // black pixel
	for (int i = 4; i < nb; i += 8) data[i + 3] = 0;  // transparent pixel
	IdStripeTexture = CORECreateTextureFromRGBA(data, 32, 32);
}

int G2D::InitSprite( V2 & Size, const string & Sprite)
{
	int LargPix = Sprite.find(']') + 1;
	Size.x = LargPix;
	int nb = Sprite.length();
	if (nb %  LargPix   != 0)
		cout << "Erreur taille texture/string incorrecte";

	int HautPix = nb / LargPix;
	Size.y = HautPix;


	int pos = 0;
	unsigned char data[128 * 128];

	for (int y = HautPix - 1; y >= 0; y--)
		for (int x = 0; x < LargPix; x++)
		{
			int  p = x + y * LargPix;
			char c = Sprite[p];
			for (int t = 0; t < 4; t++)
				data[pos++] = LetterToColor[c][t];
		}

	int IdTexture =  CORECreateTextureFromRGBA(data, LargPix, HautPix);
	return IdTexture;
}
  

 


/////////////////////////////////////////////////////////////
//
//	    RectWithTexture
//
/////////////////////////////////////////////////////////////



void G2D::DrawRectWithTexture(int texture, V2 pos, V2 size, float angleDeg )
{
	glLineWidth(0);

	float w = size.x;
	float h = size.y;
	const double M_PI = 3.14159265358979323846;
	glColor4ub(0, 0, 0, 128);

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glBindTexture(GL_TEXTURE_2D, texture);
	glEnable(GL_TEXTURE_2D);
	

	//angleDeg = 45;
	float angleRad = angleDeg / 180 * M_PI;

	float rX = cos(angleRad) * w - sin(angleRad) * h;
	float rY = sin(angleRad) * w + cos(angleRad) * h;

	float cx = pos.x + w / 2 - rX / 2;
	float cY = pos.y + h / 2 - rY / 2;

	glPushMatrix();
	glTranslatef(cx, cY, 0.0);
	glRotatef(angleDeg, 0.0, 0.0, 1.0);

	glBegin(GL_QUADS);
	glTexCoord2f(0.0, 0.0); glVertex3f(0, 0, 0.0f);
	glTexCoord2f(0.0, 1.0); glVertex3f(0, h, 0.0f);
	glTexCoord2f(1.0, 1.0); glVertex3f(w, h, 0.0f);
	glTexCoord2f(1.0, 0.0); glVertex3f(w, 0, 0.0f);
	glEnd();
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
}

/////////////////////////////////////////////////////////////
//
//	    Geometry
//
/////////////////////////////////////////////////////////////


void G2D::DrawRectangle(V2 P1, V2 Size, Color c, bool fill)
{
	glLineWidth(1);
	glColor4d(c.R, c.G, c.B, c.A);
	glDisable(GL_TEXTURE_2D);
	
	if (fill)   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	else    	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glRecti( (int) P1.x, (int)P1.y, (int)(P1.x+Size.x), (int)(P1.y+Size.y));
	 
}

void G2D::DrawLine(V2 P1, V2 P2, Color c )
{
	glLineWidth(1);
	glColor4d(c.R, c.G, c.B, c.A);

	glBegin(GL_LINES);
	glVertex2f(P1.x,P1.y);
	glVertex2f(P2.x, P2.y);
	glEnd();
}

void G2D::DrawPolygon(vector<V2> & PointList, Color c, bool fill)
{
	 
	if (fill) glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	else glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glDisable(GL_TEXTURE_2D);

	glLineWidth(1);
	glColor4d(c.R, c.G, c.B, c.A);
	
	glBegin(GL_POLYGON);
	for (V2 P : PointList)
	   glVertex2f(P.x, P.y);
	glEnd();
}

/////////////////////////////////////////////////////////////
//
//	    Font
//
/////////////////////////////////////////////////////////////


void DrawString(V2 pos, string text, float fontSize, float thickness, Color c, bool FontMono)
{
	glColor4f(c.R, c.G, c.B, c.A);

	 

	// EPAISSEUR de la font
	glLineWidth(thickness);

	glPushMatrix();
	glTranslatef(pos.x, pos.y, 0);
	glScalef(1 / 152.38, 1 / 152.38, 1 / 152.38);
	glScalef(fontSize, fontSize, fontSize);
	const char * cc = text.c_str();
	for (char* p = (char*) cc; *p; p++)
	{
		if (FontMono) glutStrokeCharacter(GLUT_STROKE_MONO_ROMAN, *p);
		else          glutStrokeCharacter(GLUT_STROKE_ROMAN, *p);
	}
	glPopMatrix();
	glLineWidth(1);
}

void G2D::Show()
{
	glFlush(); // single buffer
	glutSwapBuffers();  // double buffer
}

void G2D::DrawStringFontMono(V2 pos, string text, float fontSize, float thickness, Color c)
{
	DrawString(pos, text, fontSize, thickness, c, true);
}

void G2D::DrawStringFontRoman(V2 pos, string text, float fontSize, float thickness, Color c)
{
	DrawString(pos, text, fontSize, thickness, c, false);
}

/////////////////////////////////////////////////////////////
//
//	    Key Management
//
/////////////////////////////////////////////////////////////
 


Key keyTable[256];
Key specialKeyTable[256];

void InitKeys()
{
	for (int i = 0; i < 256; i++) keyTable[i] = specialKeyTable[i] = Key::Unknown;

	keyTable['A'] = keyTable['a'] = Key::A;
	keyTable['B'] = keyTable['b'] = Key::B;
	keyTable['C'] = keyTable['c'] = Key::C;
	keyTable['D'] = keyTable['d'] = Key::D;
	keyTable['E'] = keyTable['e'] = Key::E;
	keyTable['F'] = keyTable['f'] = Key::F;
	keyTable['G'] = keyTable['g'] = Key::G;
	keyTable['H'] = keyTable['h'] = Key::H;
	keyTable['I'] = keyTable['i'] = Key::I;
	keyTable['J'] = keyTable['j'] = Key::J;
	keyTable['K'] = keyTable['k'] = Key::K;
	keyTable['L'] = keyTable['l'] = Key::L;
	keyTable['M'] = keyTable['m'] = Key::M;
	keyTable['N'] = keyTable['n'] = Key::N;
	keyTable['O'] = keyTable['o'] = Key::O;
	keyTable['P'] = keyTable['p'] = Key::P;
	keyTable['Q'] = keyTable['q'] = Key::Q;
	keyTable['R'] = keyTable['r'] = Key::R;
	keyTable['S'] = keyTable['s'] = Key::S;
	keyTable['T'] = keyTable['t'] = Key::T;
	keyTable['U'] = keyTable['u'] = Key::U;
	keyTable['V'] = keyTable['v'] = Key::V;
	keyTable['W'] = keyTable['w'] = Key::W;
	keyTable['X'] = keyTable['x'] = Key::X;
	keyTable['Y'] = keyTable['y'] = Key::Y;
	keyTable['Z'] = keyTable['z'] = Key::Z;
	keyTable['\r'] = Key::ENTER;
	 
	specialKeyTable[GLUT_KEY_LEFT]  = Key::LEFT;
	specialKeyTable[GLUT_KEY_DOWN]  = Key::DOWN;
	specialKeyTable[GLUT_KEY_RIGHT] = Key::RIGHT;
	specialKeyTable[GLUT_KEY_UP]    = Key::UP;
	specialKeyTable[GLUT_KEY_F1]    = Key::F1;
	specialKeyTable[GLUT_KEY_F2]    = Key::F2;
	specialKeyTable[GLUT_KEY_F3]    = Key::F3;
	specialKeyTable[GLUT_KEY_F4]    = Key::F4;
	specialKeyTable[GLUT_KEY_F5]    = Key::F5;
	specialKeyTable[GLUT_KEY_F6]    = Key::F6;
	specialKeyTable[GLUT_KEY_F7]    = Key::F7;
	specialKeyTable[GLUT_KEY_F8]    = Key::F8;
	specialKeyTable[GLUT_KEY_F9]    = Key::F9;
	specialKeyTable[GLUT_KEY_F10]   = Key::F10;
	specialKeyTable[GLUT_KEY_F11]   = Key::F11;
	specialKeyTable[GLUT_KEY_F12]   = Key::F12;
}

map<Key, bool> ActiveKeys;
  
void keyboard(unsigned char key, int x, int y)   
{   
	if (key == 0x1b) exit(0);
	ActiveKeys[keyTable[key]] = true;   
}
void keyUp(unsigned char key, int x, int y)       {   ActiveKeys[keyTable[key]] = false;  }
 

void keySpecial(int key, int x, int y)            {    ActiveKeys[specialKeyTable[key]] = true; }
void keySpecialUp(int key, int x, int y)          {    ActiveKeys[specialKeyTable[key]] = false; }

bool G2D::IsKeyPressed(Key k)
{
	if (ActiveKeys.find(k) == ActiveKeys.end())      return false;
	return ActiveKeys[k];
}


/////////////////////////////////////////////////////////////
//
//	    window size
//
/////////////////////////////////////////////////////////////


int Wwidth, Wheight;

void reshape(int w, int h)
{
	glutReshapeWindow(Wwidth, Wheight);
}

/////////////////////////////////////////////////////////////
//
//	    Mouse
//
/////////////////////////////////////////////////////////////

struct mouseEvent { int x; int y; bool clicked; };
 
mouseEvent mousePress[3];

void G2D::GetMouseButtonInfo(int idButton, bool & clicked, int & x, int & y)
{
	if (mousePress[idButton].clicked == false) { clicked = false; return; }
	// read and erase
	clicked = true;
	x = mousePress[idButton].x;
	y = mousePress[idButton].y;
	mousePress[idButton].clicked = false;
}

void mouse(int button, int state, int x, int y)
{
	mouseEvent M;
	M.x = x;
	M.y = Wheight-y;
	M.clicked = true;
	if (state == GLUT_DOWN)
		mousePress[button] = M;

	//  GLUT_LEFT_BUTTON    GLUT_MIDDLE_BUTTON 

	cout << x << "   " << y << "   " << state << endl;
}



/////////////////////////////////////////////////////////////
//
//	    Init
//
/////////////////////////////////////////////////////////////

void G2D::ClearScreen(Color c)
{
	glClearColor(c.R,c.G,c.B,c.A);
	glClear(GL_COLOR_BUFFER_BIT);
}
 

void G2D::InitWindow(int & argc, char** & argv, V2 WindowSize, V2 WindowStartPos, std::string name)
{
	glutInit(&argc, argv);

	Wwidth  = WindowSize.x;
	Wheight = WindowSize.y;

	glutInitWindowPosition(WindowStartPos.x, WindowStartPos.y);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(Wwidth, Wheight);
	glutCreateWindow(name.c_str());

 
	InitColors();
	InitStripeTexture();
}

void(*_animCallBack)(void);

typedef std::chrono::high_resolution_clock Time;
typedef std::chrono::duration<double> duration;

auto T0             = Time::now();
auto TPreviousFrame = Time::now();
auto TNewFrame      = Time::now();

bool FirstFrame = true;

void anim(int value)
{
	if (FirstFrame)
	{
		T0             = Time::now();   // dmarrage du jeu
		TPreviousFrame = Time::now();
		TNewFrame      = Time::now();
		FirstFrame     = false;
	}
	else
	{
		TPreviousFrame = TNewFrame;
		TNewFrame = Time::now();
	}
	_animCallBack();
	glutPostRedisplay();
	glutTimerFunc(10, anim, value);	// max 20 FPS
	
}

 
double G2D::ElapsedTimeFromStartSeconds()
{
	duration elapsed = TNewFrame - T0;
	return elapsed.count()*0.001;
}

double G2D::ElapsedTimeFromLastCallbackSeconds()
{
	duration elapsed = TNewFrame - TPreviousFrame;
	return elapsed.count()*0.001;
}






void G2D::Run(void logic(), void render())
{
	glLoadIdentity();
	glOrtho(0, Wwidth, 0, Wheight, -1, 10);    // donne l'echelle x et y
	glMatrixMode(GL_MODELVIEW);
	glEnable(GL_TEXTURE_2D);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	glEnable(GL_LINE_SMOOTH);
	glLineWidth(1);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glDisable(GL_LIGHTING);
	glDisable(GL_CULL_FACE);
	// glEnable(GL_POLYGON_SMOOTH);  viter rendu erron

	// Init 	
	InitKeys();


	// callbacks
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutKeyboardUpFunc(keyUp);
	glutSpecialFunc(keySpecial);
	glutSpecialUpFunc(keySpecialUp);
	glutMouseFunc(mouse);


	// logic
	_animCallBack = logic;
	glutDisplayFunc(render);            // fonction appele lors d'un repaint
	glutTimerFunc(100, anim, 0);         // callback de timer avec id = 0, toutes les 50m
	glutMainLoop();
}